/*
  Freelancer .FL Savegame encode/decoder

  Credits to Sherlog <sherlog@t-online.de> for finding out the algorithm

  (c) 2003 by Jor <flcodec@jors.net>
  Tweaked by Jason Hood <jadoxa@yahoo.com.au>, 31 May, 2007.

  This is free software. Permission to copy, store and use granted as long
  as this copyright note remains intact.

  Virtually rewritten by Jason Hood, 27 to 29 July, 2009.

  v2.00, 6 to 10 August, 2009.
  * line up the group reputations, tweaked the house display;
  - recognise locked_gate within [locked_gates];
  + -N will display the first 72 characters of log messages;
  + -[nN] will convert the seconds to days/hours/minutes/seconds;
  - fixed numeric nicknames;
  + added -D option to describe each file;
  + recognise mission sections;
  * process multiple input files, using -o to create an output file.

  v2.01, 15 October, 2010:
  * recognise base_equip & base_cargo;
  * use 11 columns for house reputation (-0.00123456);
  * when piping multiple files, add the file name as a comment;
  * add blank lines after the last multiplayer visit entry & rumors.
*/

#define PVERS "2.01"
#define PDATE "15 October, 2010"

#include "common.h"
#include "hashdb.h"
#include <cerrno>
#include <io.h>
#include <fcntl.h>

Common common( true );


void usage( LPCSTR prog )
{
  printf(

  "FlCodec by Jason Hood <jadoxa@yahoo.com.au>.\n"
  "Algorithm found by Sherlog; original program by Jor.\n"
  "Version " PVERS " (" PDATE ").  Freeware.\n"
  "http://freelancer.adoxa.cjb.net/\n"
  "\n"
  "Decode or encode Freelancer savegames.\n"
  "\n"
  "%s [-adehlnN] [-f DBFILE] [-o OUTFILE] INFILE...\n"
  "%s -D[al] [file...]\n"
  "\n"
  "-a\tadd the single player Auto Save to the file list\n"
  "-d\tdecode INFILE to INFILE.ini or standard output\n"
  "-e\tencode INFILE to INFILE.fl or standard output\n"
  "-f\tuse DBFILE instead of \"hashes.dat\"\n"
  "-h\tconvert nicknames to hashes (.txt to .ini/.fl)\n"
  "-l\tadd the latest single player savegame to the file list\n"
  "-n\tconvert hashes to nicknames (.fl/.ini to .txt)\n"
  "-N\tconvert hashes/nicknames to names (.fl/.ini/.txt to .lst)\n"
  "-o\twrite a single INFILE to OUTFILE\n"
  "-D\tdisplay description and timestamp of each FILE\n"
  "\n"
  "If the option is missing, INFILE will be decoded if it starts with the\n"
  "encryption signature, encoded otherwise.\n"
  "If using INFILE as output, the existing extension will be replaced\n"
  "(eg: AutoSave.fl <--> AutoSave.ini).\n"
  "Hash conversion requires CreateID's database.  Nicknames must be converted\n"
  "back to hashes for the information to be correctly loaded.  Names cannot be\n"
  "converted back again, nor will such files be described by -D.\n"
  "If no FILE is given, describe each single player game.  Games will be sorted\n"
  "the same as Freelancer, newest first.\n"

  , prog, prog );
}


void crypt( char* buff, int len )
{
  /* Very Secret Key - this is Microsoft Security In Action[tm] */
  static const char gene[] = "Gene";
  int i;

  for (i = 0; i < len; ++i)
    buff[i] ^= (gene[i & 3] + i) | 0x80;
}


// Read the file named IFILE into memory, returning the pointer to memory and
// storing the length in LEN.  If the file is encrypted, it is decrypted and FL
// is set true, otherwise false.  Does not return if anything went wrong.
char* read_file( LPCSTR ifile, int& len, bool& fl )
{
  int	ifd, rc;
  char* mem;

  ifd = open( ifile, O_RDONLY );
  if (ifd == -1)
  {
    fprintf( stderr, "Error: open(%s) failed: %s\n", ifile, strerror( errno ) );
    exit( 1 );
  }

  len = lseek( ifd, 0, SEEK_END );
  lseek( ifd, 0, SEEK_SET );

  int sig;
  read( ifd, &sig, 4 );
  if (sig == '1SLF')
  {
    fl = true;
    len -= 4;
  }
  else
  {
    fl = false;
    lseek( ifd, 0, SEEK_SET );
  }

  mem = (char*)malloc( len );
  if (mem == NULL)
  {
    fprintf( stderr, "Error: malloc(%d) failed: %s\n", len, strerror( errno ) );
    exit( 1 );
  }

  rc = read( ifd, mem, len );
  if (rc != len)
  {
    fprintf( stderr, "Error: read() failed: %s\n", strerror( errno ) );
    exit( 1 );
  }

  close( ifd );

  if (fl)
    crypt( mem, len );

  return mem;
}


// Write LEN bytes at MEM to the file named OFILE (or standard output, if OFILE
// is empty).  If FL is true, the FLS1 signature is written and the memory is
// encrypted before being written.  Does not return if anything went wrong.
void write_file( LPCSTR ofile, char* mem, int len, bool fl )
{
  int ofd, rc;

  ofd = (*ofile) ? open( ofile, O_CREAT | O_TRUNC | O_WRONLY, 0640 ) : 1;
  if (ofd == -1)
  {
    fprintf( stderr, "Error: open(%s) failed: %s\n", ofile, strerror( errno ) );
    exit( 1 );
  }

  if (fl)
  {
    /* write magic token */
    rc = write( ofd, "FLS1", 4 );
    if (rc != 4)
    {
      fprintf( stderr, "Error: write() failed: %s\n", strerror( errno ) );
      exit( 1 );
    }
    crypt( mem, len );
  }

  rc = write( ofd, mem, len );
  if (rc != len)
  {
    fprintf( stderr, "Error: write() failed: %s\n", strerror( errno ) );
    exit( 1 );
  }

  if (ofd != 1)
    close( ofd );
}


enum { NOTHING, NICK, NAME, HASH };
int    option;
HashDB hash;
FILE*  file;

void   process( INI_Reader& );
UINT   get_id( INI_Reader&, int );
UINT   get_faction( INI_Reader&, int );
LPCSTR xlat( UINT );
void   xlat_secs( LPCSTR );
LPCSTR get_string( INI_Reader&, int, LPSTR, int );
void   describe_game( LPCSTR );
void   dump_games();


int main( int argc, char* argv[] )
{
  char	outfile[MAX_PATH];
  char* oname = NULL;

  _fmode = O_BINARY;

  if (argc == 1 ||
      ((argv[1][0] == '/' || argv[1][0] == '-') && argv[1][1] == '?') ||
      strcmp( argv[1], "--help" ) == 0)
  {
    usage( argv[0] );
    return 0;
  }

  LPCSTR dbname = "hashes.dat";
  enum { DETERMINE, DECODE, ENCODE, DESCRIBE };
  int command = DETERMINE;
  bool autosave = false, latest = false;

  while (*++argv && **argv == '-')
  {
    while (*++*argv)
    {
      switch (**argv)
      {
	case 'a': autosave = true;     break;
	case 'D': command  = DESCRIBE; break;
	case 'd': command  = DECODE;   break;
	case 'e': command  = ENCODE;   break;
	case 'h': option   = HASH;     break;
	case 'l': latest   = true;     break;
	case 'n': option   = NICK;     break;
	case 'N': option   = NAME;     break;
	case 'f':
	case 'o':
	  if (!argv[0][1] && !argv[1])
	  {
	    fprintf( stderr, "Expecting file name for -%c.\n", **argv );
	    return 1;
	  }
	  {
	  char	opt = **argv;
	  char* fn = (argv[0][1]) ? argv[0] + 1 : *++argv;
	  if (opt == 'f')
	    dbname = fn;
	  else
	    oname  = fn;
	  }
	goto next_arg;
	default:
	  fprintf( stderr, "Unknown option '%c'.\n", **argv );
	  return 1;
      }
    }
    next_arg: ;
  }

  char autosavefl[MAX_PATH], latestfl[MAX_PATH];
  if (autosave || latest)
  {
    char sp[MAX_PATH];
    GetUserDataPath( sp );
    strcat( sp, "\\Accts\\SinglePlayer\\" );
    if (latest)
    {
      FILETIME ft = { 0, 0 };
      WIN32_FIND_DATA fd;
      strcat( sp, "*.fl" );
      HANDLE find = FindFirstFile( sp, &fd );
      if (find != INVALID_HANDLE_VALUE)
      {
	*(LPSTR)get_file_name( sp ) = '\0';
	do
	{
	  if (CompareFileTime( &fd.ftLastWriteTime, &ft ) > 0 &&
	      stricmp( fd.cFileName, "AutoSave.fl" ) != 0)
	  {
	    ft = fd.ftLastWriteTime;
	    sprintf( latestfl, "%s%s", sp, fd.cFileName );
	  }
	} while (FindNextFile( find, &fd ));
	FindClose( find );
      }
      if (ft.dwLowDateTime || ft.dwHighDateTime)
	*--argv = latestfl;
    }
    if (autosave)
    {
      sprintf( autosavefl, "%s%s", sp, "AutoSave.fl" );
      *--argv = autosavefl;
    }
  }

  if (command == DESCRIBE)
  {
    if (!*argv)
    {
      char sp[MAX_PATH];
      GetUserDataPath( sp );
      SetCurrentDirectory( sp );
      SetCurrentDirectory( "Accts\\SinglePlayer" );
      WIN32_FIND_DATA fd;
      HANDLE find = FindFirstFile( "*.fl", &fd );
      if (find != INVALID_HANDLE_VALUE)
      do
	describe_game( fd.cFileName );
      while (FindNextFile( find, &fd ));
      FindClose( find );
    }
    else
      common.process_args( argv, describe_game, NULL );
    dump_games();
    return 0;
  }

  if (!*argv)
  {
    fputs( "File name required.\n", stderr );
    return 1;
  }
  if (argv[1] && oname)
  {
    fputs( "Only one input file allowed when specifying an output file.\n",
	   stderr );
    return 1;
  }

  if ((option == NICK || option == NAME) && !hash.open( dbname ))
    return 1;

  INI_Reader ini;
  bool multiple = (argv[1] != NULL), first = true;
  for (; *argv != NULL; ++argv)
  {
    int   len;
    bool  fl;
    char* mem = read_file( *argv, len, fl );

    int cmd = command;
    if (option == NOTHING && cmd == DETERMINE)
      cmd = (fl) ? DECODE : ENCODE;

    if (oname)
    {
      strncpy( outfile, oname, sizeof(outfile) - 1 );
      outfile[sizeof(outfile)-1] = '\0';
    }
    else if (!isatty( 1 ))
    {
      *outfile = '\0';
    }
    else
    {
      char* ext = strrchr( *argv, '.' );
      if (!ext)
	ext = strchr( *argv, '\0' );
      sprintf( outfile, "%.*s.%s", ext - *argv, *argv,
	       (cmd == ENCODE) ? "fl" :
	       (option == NOTHING || option == HASH) ? "ini" :
	       (option == NICK) ? "txt" :
	     /*(option == NAME)*/ "lst" );
    }

    if (option == NOTHING)
    {
      if (cmd == ENCODE && fl)
	fprintf( stderr, "File \"%s\" is already encoded.\n", *argv );
      else if (command == DECODE && !fl)
	fprintf( stderr, "File \"%s\" is already decoded.\n", *argv );
      else
      {
	if (!*outfile && multiple)
	{
	  if (!first)
	    write( 1, "\n", 1 );
	  first = false;
	  write( 1, "; ", 2 );
	  write( 1, *argv, strlen( *argv ) );
	  write( 1, "\n", 1 );
	}
	write_file( outfile, mem, len, !fl );
      }
      free( mem );
      continue;
    }

    if (!ini.open_memory( mem, len ))
    {
      fprintf( stderr, "Error: open_memory(%s) failed.\n", *argv );
      return 1;
    }
    file = (*outfile) ? fopen( outfile, "wb" ) : stdout;
    if (!file)
    {
      fprintf( stderr, "Error: open(%s) failed: %s\n",
		       outfile, strerror( errno ) );
      return 1;
    }

    if (!*outfile)
    {
      if (multiple)
      {
	if (!first)
	  putchar( '\n' );
	first = false;
	printf( "; %s\n", *argv );
      }
    }

    process( ini );
    if (*outfile)
      fclose( file );

    if (cmd == ENCODE)
    {
      free( mem );
      mem = read_file( outfile, len, fl );
      write_file( outfile, mem, len, true );
    }

    ini.close();
    free( mem );
  }

  return 0;
}


void process( INI_Reader& ini )
{
  LPCSTR val;
  UINT	 id;
  char	 num[16];

  int first_hdr = 1;
  while (ini.read_header())
  {
    fprintf( file, "\n[%s]\n" + first_hdr, ini.get_header_ptr() );
    first_hdr = 0;

    if (ini.is_header( "Player" ))
    {
      // Separate groups of entries with blank lines.
      bool in_group = false;
      while (ini.read_value())
      {
	int no_nl = 1;
	if (ini.is_value( "interface" ))
	{
	  no_nl = 0;
	}
	else if (ini.is_value( "voice" )  ||
		 ini.is_value( "system" ) ||
		 ini.is_value( "ship_archetype" ))
	{
	  no_nl = 0;
	}
	else if (ini.is_value( "house" ) ||
		 ini.is_value( "wg" )    ||
		 ini.is_value( "log" ))
	{
	  if (!in_group)
	  {
	    in_group = true;
	    no_nl = 0;
	  }
	}
	else if (in_group)
	{
	  in_group = false;
	  no_nl = 0;
	}
	fprintf( file, "\n%s = " + no_nl, ini.get_name_ptr() );
	if (option == NAME)
	{
	  if (ini.is_value( "description" ) ||
	      ini.is_value( "name" ))
	  {
	    val = ini.get_value_string();
	    WCHAR wval[1024];
	    HexToWide( val, wval );
	    fprintf( file, "%S\n", wval );
	    continue;
	  }
	  if (ini.is_value( "descrip_strid" ))
	  {
	    char desc[1024];
	    common.GetString( ini.get_value_int( 0 ), desc, sizeof(desc) );
	    fprintf( file, "%s\n", desc );
	    continue;
	  }
	  if (ini.is_value( "tstamp" ))
	  {
	    FILETIME ft;
	    ft.dwHighDateTime = ini.get_value_int( 0 );
	    ft.dwLowDateTime  = ini.get_value_int( 1 );
	    fprintf( file, "%S\n", filetime_to_wstring( ft ) );
	    continue;
	  }
	  if (ini.is_value( "system" ) ||
	      ini.is_value( "base" )   ||
	      ini.is_value( "rep_group" ))
	  {
	    fprintf( file, "%s\n", xlat( get_id( ini, 0 ) ) );
	    continue;
	  }
	  if (ini.is_value( "log" ))
	  {
	    char temp[73];
	    fprintf( file, "%s\n", get_string( ini, 0, temp, sizeof(temp) ) );
	    continue;
	  }
	}
	// Since I'm here, make the reputations a little nicer on the eye.
	if (ini.is_value( "house" ))
	{
	  // Number conversion uses the same storage as string.
	  float rep = ini.get_value_float( 0 );
	  if (option == NAME)
	  {
	    val = xlat( get_id( ini, 1 ) );
	    fprintf( file, "% -11g  %-16s  %s", rep,
		     (rep <= -0.6f) ? "Hostile" :
		     (rep <= -0.2f) ? "Neutral Hostile" :
		     (rep <   0.2f) ? "Neutral" :
		     (rep <   0.6f) ? "Neutral Friendly" :
				      "Friendly", val );
	  }
	  else
	  {
	    sprintf( num, "% g,", rep );
	    fprintf( file, "%-12s %s", num, ini.get_value_string( 1 ) );
	  }
	}
	else if (ini.is_value( "equip" ) || ini.is_value( "base_equip" ) ||
		 ini.is_value( "cargo" ) || ini.is_value( "base_cargo" ))
	{
	  // Skip over the first string, copying the rest verbatim.
	  val = strchr( ini.get_value_string(), ',' );
	  if (val == NULL)
	    val = "";
	  fprintf( file, "%s%s", xlat( get_id( ini, 0 ) ), val );
	}
	else if (ini.is_value( "visit" ))
	{
	  int flag = ini.get_value_int( 1 );
	  id = (flag == 65) ? get_faction( ini, 0 ) : get_id( ini, 0 );
	  fprintf( file, "%s, %d", xlat( id ), flag );
	}
	else if (ini.is_value( "location" ))
	{
	  id = ini.get_value_int( 0 );
	  if (!id)
	    id = get_hash( ini.get_value_string() );
	  fputs( xlat( id ), file );
	}
	else if (ini.is_value( "com_body" )      ||
		 ini.is_value( "com_head" )      ||
		 ini.is_value( "com_lefthand" )  ||
		 ini.is_value( "com_righthand" ) ||
		 ini.is_value( "body" )          ||
		 ini.is_value( "head" )          ||
		 ini.is_value( "lefthand" )      ||
		 ini.is_value( "righthand" )     ||
		 ini.is_value( "location" )      ||
		 ini.is_value( "ship_archetype" ))
	{
	  fputs( xlat( get_id( ini, 0 ) ), file );
	}
	else
	{
	  fputs( ini.get_value_string(), file );
	}
	fputc( '\n', file );
      }
    }
    else if (ini.is_header( "mPlayer" ))
    {
      bool in_group = false;
      while (ini.read_value())
      {
	int no_nl = 1;
	if (ini.is_value( "total_cash_earned" ) ||
	    ini.is_value( "can_dock" ))
	{
	  no_nl = 0;
	  in_group = false;
	}
	else if (ini.is_value( "vnpc" )             ||
		 ini.is_value( "ship_type_killed" ) ||
		 ini.is_value( "base_visited" ))
	{
	  if (!in_group)
	  {
	    in_group = true;
	    no_nl = 0;
	  }
	}
	else if (in_group)
	{
	  in_group = false;
	  no_nl = 0;
	}
	fprintf( file, "\n%s = " + no_nl, ini.get_name_ptr() );
	if (ini.is_value( "vnpc" ))
	{
	  fprintf( file, "%s, %s, %d, %d", xlat( get_id( ini, 0 ) ),
					   xlat( get_id( ini, 1 ) ),
					   ini.get_value_int( 2 ),
					   ini.get_value_int( 3 ) );
	}
	else if (ini.is_value( "ship_type_killed" ))
	{
	  fprintf( file, "%s, %d", xlat( get_id( ini, 0 ) ),
				   ini.get_value_int( 1 ) );
	}
	else if (ini.is_value( "total_time_played" ))
	{
	  xlat_secs( ini.get_value_string() );
	  // It also adds a blank before the first sys_visited.
	}
	else if (ini.is_value( "locked_gate" )  ||
		 ini.is_value( "sys_visited" )  ||
		 ini.is_value( "base_visited" ) ||
		 ini.is_value( "holes_visited"))
	{
	  fputs( xlat( get_id( ini, 0 ) ), file );
	}
	else
	{
	  fputs( ini.get_value_string(), file );
	}
	fputc( '\n', file );
      }
    }
    else if (ini.is_header( "RandomMission" ))
    {
      while (ini.read_value())
      {
	if (ini.is_value( "offer_base" ) ||
	    ini.is_value( "dest_sys" )   ||
	    ini.is_value( "objective_zone" ))
	{
	  val = xlat( get_id( ini, 0 ) );
	}
	else if (ini.is_value( "offer_group" ) ||
		 ini.is_value( "enemy_group" ))
	{
	  val = xlat( get_faction( ini, 0 ) );
	}
	else
	  val = ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (ini.is_header( "MissionState" ))
    {
      while (ini.read_value())
      {
	if (option == NAME && (ini.is_value( "msn_offer" ) ||
			       ini.is_value( "msn_title" )))
	{
	  char buf[67];
	  val = get_string( ini, 0, buf, sizeof(buf) );
	}
	else
	  val = ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (ini.is_header( "TriggerSave" ))
    {
      while (ini.read_value())
      {
	val = (ini.is_value( "trigger" )) ? xlat( get_id( ini, 0 ) )
					  : ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (ini.is_header( "MsnShipSave" ))
    {
      while (ini.read_value())
      {
	val = (ini.is_value( "npc" ) ||
	       ini.is_value( "label" )) ? xlat( get_id( ini, 0 ) )
					: ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (ini.is_header( "MsnWingSave" ))
    {
      while (ini.read_value())
      {
	val = (ini.is_value( "formation" ) ||
	       ini.is_value( "arrival_obj" )) ? xlat( get_id( ini, 0 ) )
					      : ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (option == NAME && ini.is_header( "MObjective" ))
    {
      while (ini.read_value())
      {
	val = (ini.is_value( "Dock" )) ? xlat( get_id( ini, 0 ) )
				       : ini.get_value_string();
	fprintf( file, "%s = %s\n", ini.get_name_ptr(), val );
      }
    }
    else if (ini.is_header( "NNObjective" ))
    {
      while (ini.read_value())
      {
	fprintf( file, "%s = ", ini.get_name_ptr() );
	if (ini.is_value( "nickname" ))
	{
	  val = ini.get_value_string();
	  if (*val != '0' || val[1] != '\0')
	    val = xlat( get_id( ini, 0 ) );
	  fputs( val, file );
	}
	else if (option == NAME && ini.is_value( "type" ))
	{
	  char obj[56];
	  id = ini.get_value_int( 0 );
	  if (id == 2 || id == 4)
	  {
	    fprintf( file, "%d, %s, %s", id,
			   xlat( get_id( ini, 1 ) ),
			   get_string( ini, 2, obj, sizeof(obj) ) );
	    if (id == 4)
	      fprintf( file, ", %s", xlat( get_id( ini, 7 ) ) );
	  }
	  else if (id == 3 || id == 5)
	  {
	    fprintf( file, "%d, %s", id,
			   get_string( ini, 1, obj, sizeof(obj) ) );
	    if (id == 5)
	      fprintf( file, ", %s", xlat( get_id( ini, 2 ) ) );
	  }
	  else
	    fputs( ini.get_value_string(), file );
	}
	else
	  fputs( ini.get_value_string(), file );
	fputc( '\n', file );
      }
    }
    else if (ini.is_header( "RandomEncounter" ))
    {
      while (ini.read_value())
      {
	fprintf( file, "%s = ", ini.get_name_ptr() );
	if (ini.is_value( "formation" ))
	{
	  val = ini.get_value_string( 0 );
	  id = (*val == '0' && val[1] == 'x')
	       ? strtoul( ini.get_value_string( 0 ), NULL, 0 )
	       : get_id( ini, 0 );
	  fprintf( file, "%s, %f, %d, %d, %s, %f",
			 xlat( id ),
			 ini.get_value_float( 1 ),
			 ini.get_value_int( 2 ),
			 ini.get_value_int( 3 ),
			 xlat( get_id( ini, 4 ) ),
			 ini.get_value_float( 5 ) );
	}
	else if (ini.is_value( "attacker_rep" ))
	  fputs( xlat( get_faction( ini, 0 ) ), file );
	else
	  fputs( ini.get_value_string(), file );
	fputc( '\n', file );
      }
    }
    else if (ini.is_header( "Time" ))
    {
      ini.read_value();
      fprintf( file, "%s = ", ini.get_name_ptr() );
      xlat_secs( ini.get_value_string() );
    }
    // Align the group reputations.
    else if (ini.is_header( "Group" ))
    {
      // Let's assume nickname is first.
      ini.read_value();
      fprintf( file, "%s\n", ini.get_line_ptr() );
      // Now it's all rep, and let's assume 2 decimals.
      while (ini.read_value())
      {
	fprintf( file, "%s = ", ini.get_name_ptr() );
	val = ini.get_value_string( 0 );
	if (*val == '-')
	{
	  ++val;
	  fputc( '-', file );
	}
	else
	  fputc( ' ', file );
	sprintf( num, "%s,", val );
	fprintf( file, "%-5s %s\n", num, ini.get_value_string( 1 ) );
      }
    }
    else if (ini.is_header( "locked_gates" ))
    {
      bool fnd_locked = false;
      while (ini.read_value())
      {
	if (!fnd_locked && ini.is_value( "locked_gate" ))
	{
	  fnd_locked = true;
	  fputc( '\n', file );
	}
	fprintf( file, "%s = %s\n", ini.get_name_ptr(),
		       xlat( get_id( ini, 0 ) ) );
      }
    }
    else
    {
      while (ini.read_value())
	fprintf( file, "%s\n", ini.get_line_ptr() );
    }
  }
}


UINT get_id( INI_Reader& ini, int idx )
{
  UINT id = ini.get_value_int( idx );
  if (!id || !(id & 0x80000000))
    id = CreateID( ini.get_value_string( idx ) );
  return id;
}


UINT get_faction( INI_Reader& ini, int idx )
{
  UINT id = ini.get_value_int( idx );
  if (!id)
    id = MakeID( ini.get_value_string( idx ) );
  return id;
}


LPCSTR xlat( UINT id )
{
  static char num[2][12];
  static int  idx;

  if (hash.find( id ))
  {
    if (option == NICK || (option == NAME && !*hash.name()))
      return hash.nick();
    if (option == NAME)
      return hash.name();
  }
  idx ^= 1;
  sprintf( num[idx], "%u", id );
  return num[idx];
}


void xlat_secs( LPCSTR secstr )
{
  int	days, hrs, mins;
  float secs;

  if (strchr( secstr, ':' ))
    sscanf( secstr, "%d:%d:%d:%f", &days, &hrs, &mins, &secs );
  else
  {
    secs = atof( secstr );
    int isecs = atoi( secstr );
    secs -= isecs;
    days = isecs / 86400; isecs %= 86400;
    hrs  = isecs /  3600; isecs %=  3600;
    mins = isecs /    60;
    secs += isecs % 60;
  }
  if (option == NAME)
  {
    #define PLURAL(var) var, "s" + (var == 1)
    if (days)
      fprintf( file, "%d day%s", PLURAL(days) );
    if (hrs)
      fprintf( file, ", %d hour%s" + (days ? 0 : 2), PLURAL(hrs) );
    if (mins)
      fprintf( file, ", %d minute%s" + ((days | hrs) ? 0 : 2), PLURAL(mins) );
    if (secs)
      fprintf( file, ", %g second%s" + ((days|hrs|mins) ? 0 : 2),PLURAL(secs) );
  }
  else if (option == NICK)
    fprintf( file, "%d:%02d:%02d:%02g", days, hrs, mins, secs );
  else
    fprintf( file, "%g", days * 86400 + hrs * 3600 + mins * 60 + secs );

  fputc( '\n', file );
}


LPCSTR get_string( INI_Reader& ini, int idx, LPSTR buf, int max )
{
  if (common.GetString( ini.get_value_int( idx ), buf, max, false ))
  {
    // Replace newlines with spaces.
    for (char* nl = strchr( buf, '\n' ); nl; nl = strchr( nl, '\n' ))
      *nl = ' ';
  }
  else
    strcpy( buf, ini.get_value_string( idx ) );

  return buf;
}


typedef std::pair<string, string> FileDesc;
typedef std::multimap<FILETIME, FileDesc> GameMap;
GameMap games;

// Sort the same as Freelancer, newest first.
bool std::less<FILETIME>::operator()( const FILETIME& ft1,
				      const FILETIME& ft2 ) const
{
  return (CompareFileTime( &ft1, &ft2 ) > 0);
}


void describe_game( LPCSTR name )
{
  int	len;
  bool	fl;
  char* mem = read_file( name, len, fl );

  INI_Reader ini;
  if (ini.open_memory( mem, len ) && ini.find_header( "Player" ))
  {
    FILETIME ft = { 0, 0 };
    FileDesc fd;
    fd.first  = get_file_name( name );
    fd.second = "";
    bool fnd_desc = false, fnd_tstamp = false, no_desc = false;
    while (ini.read_value())
    {
      if (ini.is_value( "description" ))
      {
	WCHAR wdesc[1024];
	HexToWide( ini.get_value_string(), wdesc );
	CHAR desc[1024];
	wcstombs( desc, wdesc, sizeof(desc) );
	fd.second = desc;
	if (fnd_tstamp)
	  break;
	fnd_desc = true;
      }
      else if (ini.is_value( "descrip_strid" ))
      {
	char desc[46];
	common.GetString( ini.get_value_int( 0 ), desc, sizeof(desc) );
	fd.second = desc;
	if (fnd_tstamp)
	  break;
	fnd_desc = true;
      }
      else if (ini.is_value( "tstamp" ))
      {
	ft.dwHighDateTime = ini.get_value_int( 0 );
	// Don't do list files.
	if (ft.dwHighDateTime < 4096)
	{
	  no_desc = true;
	  break;
	}
	ft.dwLowDateTime = ini.get_value_int( 1 );
	if (fnd_desc)
	  break;
	fnd_tstamp = true;
      }
    }
    if (!no_desc)
      games.insert( GameMap::value_type( ft, fd ) );
  }

  ini.close();
  free( mem );
}


void dump_games()
{
  GameMap::const_iterator g, end = games.end();
  for (g = games.begin(); g != end; ++g)
    printf( "%-15s%-48s%S\n", g->second.first.c_str(),
			      g->second.second.c_str(),
			      filetime_to_wstring( g->first ) );
}
